""" Utilities for interacting with the Syndetics catalog.

Syndetics provides a repository of book covers, among other things.
"""
import urllib.parse as urlparse


def higher_quality_cover(image_url: str) -> str:
    """Modify a book cover request to be higher quality.

    By default, the BiblioCommons catalog shows medium-quality images (GIF,
    187x187). There are higher-quality images available though (JPG, 400x400).

    This method returns a new URL that requests the same resource from the
    Syndetics catalog, but in higher quality. The other querystring parameters,
    such as those identifying the client or making requests for certain
    metadata, are left unmodified.

    Documentation on the URL scheme can be found in the 'Syndetics Starter Page':
    https://developers.exlibrisgroup.com/wp-content/uploads/2010/05/SyndeticsStarterDocument.pdf
    """
    parsed = urlparse.urlparse(image_url)
    params = urlparse.parse_qs(parsed.query)
    if 'isbn' not in params:
        # We might receive a placeholder image, use that instead
        return image_url

    isbn, _filename = params['isbn'][0].split('/')  # '123456789/SC.GIF'
    params['isbn'] = [f"{isbn}/LC.jpg"]  # (higher quality image)
    large_cover = parsed._replace(query=urlparse.urlencode(params, doseq=True))
    return large_cover.geturl()



import logging
import re
import urllib.parse as urlparse

logger = logging.getLogger('bibliophile')

# Expect image urls to conform to a certain scheme
GOODREADS_IMAGE_REGEX = re.compile(
    # Base URL could be any of:
    # - 'https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/(...)'
    # - 'https://images.gr-assets.com/books/(...)'
    # - 'http://compressed.photo.goodreads.com/books/(...)'
    r'/(?P<base>(images/S/compressed\.photo\.goodreads\.com/books|books))/'
    r'(?P<numeric>\d+)'  # Some numeric ID that doesn't seem to be ISBN...
    r'(?P<size>[sml])/'  # size: 'small', 'medium', or 'large'
    r'(?P<slug>\d+)'  # slug (identifier on the book's home page)
    r'((?P<resize>\._S(?P<dimension>[XY]\d+)_))?'  # optional resize (e.g. '17081913._SX98_' is 98 pixels wide)
    r'\.(?P<extension>jpg)$'
)


def higher_quality_cover(image_url: str) -> str:
    """ Modify a book cover to be higher quality. """
    parsed = urlparse.urlparse(image_url)
    if parsed.path.startswith('/assets/nophoto'):
        # No known cover for this book! Just return the "no photo" image
        return image_url

    match = GOODREADS_IMAGE_REGEX.match(parsed.path)
    if not match:
        logger.warning(
            "Goodreads image format changed! (%s) " "Returning original quality image.",
            image_url,
        )
        return image_url

    # The new path excludes:
    # - the resize group (which means we'll render full size)
    # - the original size (we'll use large automatically)
    larger_path = "/{base}/{numeric}l/{slug}.{extension}".format_map(match.groupdict())

    # Make sure to keep the original domain
    return parsed._replace(path=larger_path).geturl()



class UnstableAPIError(RuntimeError):
    """Indicates a failed assumption about an unstable API.

    This whole project is built off of undocumented APIs.
    If raised, this exception indicates that something about the API has changed,
    and the source code should be modified accordingly.
    """



import unittest

from ..syndetics import higher_quality_cover


class SyndeticsTest(unittest.TestCase):
    def test_unrecognize_images_returned_as_is(self):
        """ If there's no ISBN in the URL, we don't modify the URL. """
        img_url = 'https://secure.syndetics.com/index.aspx?fake_url'
        self.assertEqual(higher_quality_cover(img_url), img_url)

    def test_small_gif_becomes_large_jpeg(self):
        """ We transform small images to larger ones. """
        base_url = 'https://secure.syndetics.com/index.aspx'
        self.assertEqual(
            higher_quality_cover(
                f'{base_url}?isbn=9780140449266/SC.GIF&client=sfpl&type=xw12&oclc='
            ),
            f'{base_url}?isbn=9780140449266%2FLC.jpg&client=sfpl&type=xw12',
        )



import unittest

from bibliophile.bibliocommons.errors import QueryError
from bibliophile.bibliocommons.query import QueryBuilder
from bibliophile.bibliocommons.types import BookDescription
from bibliophile.goodreads import Book as GoodreadsBook

don_quixote = BookDescription.from_goodreads_book(
    GoodreadsBook(
        goodreads_id="3836",
        isbn="0142437239",
        title="Don Quixote",
        author="Miguel de Cervantes",
        description="Windmills, Sancho Panza, etc.",
        image_url="not relevant, so not a real URL",
    )
)

moby_dick = BookDescription.from_goodreads_book(
    GoodreadsBook(
        goodreads_id="153747",
        isbn=None,  # It *does* have an ISBN, but we're pretending...
        title="Moby Dick",
        author="Herman Melville",
        description="Ahab just can't let it go",
        image_url="not relevant, so not a real URL",
    )
)


class QueryBuilderTest(unittest.TestCase):
    def test_single_isbn(self):
        self.assertEqual(
            QueryBuilder.bibliocommons_query(
                books=[don_quixote], branch=None, isolanguage=None
            ),
            'identifier:(0142437239)',
        )

    def test_no_known_isbn(self):
        self.assertEqual(
            QueryBuilder.bibliocommons_query(
                books=[moby_dick], branch=None, isolanguage=None
            ),
            '(contributor:(Herman Melville) AND title:(Moby Dick) AND formatcode:(BK))',
        )

    def test_single_book_with_branch(self):
        self.assertEqual(
            QueryBuilder.bibliocommons_query(
                books=[don_quixote], branch='Fremont Branch', isolanguage=None
            ),
            '(identifier:(0142437239)) available:"Fremont Branch"',
        )

    def test_single_book_with_branch_and_language(self):
        self.assertEqual(
            QueryBuilder.bibliocommons_query(
                books=[don_quixote], branch='Fremont Branch', isolanguage='eng'
            ),
            '(identifier:(0142437239)) available:"Fremont Branch" isolanguage:"eng"',
        )

    def test_two_books(self):
        self.assertEqual(
            QueryBuilder.bibliocommons_query(
                books=[don_quixote, moby_dick],
                branch='Fremont Branch',
                isolanguage='eng',
            ),
            (
                # Don Quixote has an ISBN
                '(identifier:(0142437239)'
                ' OR '
                # Moby Dick has no ISBN - so we look for title, Author, book type
                '(contributor:(Herman Melville) AND title:(Moby Dick) AND formatcode:(BK)))'
                # Both books are queried as in English and at the Fremont Branch
                ' available:"Fremont Branch" isolanguage:"eng"'
            ),
        )

    def test_no_book_options_okay(self):
        self.assertEqual(
            QueryBuilder.bibliocommons_query(
                books=[moby_dick], branch=None, isolanguage=None, print_only=False
            ),
            '(contributor:(Herman Melville) AND title:(Moby Dick))',
        )

    def test_query_way_too_long(self):
        """ A query is only valid when it is below 900 characters. """
        lots_of_books = [
            GoodreadsBook(
                goodreads_id="123456",
                isbn=f"0000000{i:03d}",
                title="Unknown",
                author="Unknown",
                description="This won't be used in the query",
                image_url="not relevant, so not a real URL",
            )
            for i in range(50)
        ]

        # It's valid to query for a small number of books
        QueryBuilder.bibliocommons_query(
            books=lots_of_books[:10], branch=None, isolanguage=None
        )
        # Querying for too many introduces an exception
        with self.assertRaises(QueryError):
            QueryBuilder.bibliocommons_query(
                books=lots_of_books, branch=None, isolanguage=None
            )



from typing import Iterable, Optional

from .errors import QueryError
from .types import BookDescription

# BiblioCommons will ignore any query over 900 characters.
# This was experimentally derived using the ASCII character set.
# I have no idea how Unicode code points are handled (or if they are handled at all)
MAX_QUERY_LENGTH = 900


class QueryBuilder:  # pylint: disable=too-few-public-methods
    """ Construct BiblioCommons catalog queries for one or more books. """

    @staticmethod
    def _single_query(book: BookDescription, print_only: bool = True) -> str:
        """ Get query for one book - Use its ISBN (preferred) or title + author. """
        conditions = {}

        if book.isbn:
            conditions['identifier'] = book.isbn
        else:
            conditions['contributor'] = book.author
            conditions['title'] = book.title
            if print_only:
                conditions['formatcode'] = 'BK'

        rules = [f'{name}:({val})' for name, val in conditions.items()]
        query = ' AND '.join(rules)
        return f'({query})' if len(rules) > 1 else query

    @classmethod
    def _query(
        cls,
        books: Iterable[BookDescription],
        branch: Optional[str],
        isolanguage: Optional[str],
        print_only: bool,
    ) -> str:
        """Get query for "any of these books available at this branch."

        Note that this query may not actually be accepted!
        Use the public method, `bibliocommons_query()` for validation.
        """
        isbn_match = ' OR '.join(
            cls._single_query(book, print_only=print_only) for book in books
        )
        parts = []
        if branch:
            parts.append(f'available:"{branch}"')
        if isolanguage:
            parts.append(f'isolanguage:"{isolanguage}"')

        return f'({isbn_match}) {" ".join(parts)}' if parts else isbn_match

    @classmethod
    def bibliocommons_query(
        cls,
        books: Iterable[BookDescription],
        branch: Optional[str],
        isolanguage: Optional[str],
        print_only: bool = True,
    ) -> str:
        """Get query for "any of these books available [at this branch] and/or [in this language]"

        This query can be used in any Bibliocommons-driven catalog.

        If the query is invalid, an exception will be raised.
        """
        query = cls._query(
            books, branch=branch, isolanguage=isolanguage, print_only=print_only
        )

        if len(query) > MAX_QUERY_LENGTH:
            raise QueryError("BiblioCommons queries are limited to 900 chars!")
        return query



import typing
from typing import Optional


class Book(typing.NamedTuple):
    # Every Goodreads book has a slug, a string of integers which identify it.
    goodreads_id: str

    # Plain ISBN, not ISBN13
    # In rare cases, it's possible for an ISBN to be missing!
    isbn: Optional[str]

    title: str
    author: str
    description: str

    # If one exists, a book cover image
    # If no image exists, this will be a "nophoto" URL ('/assets/nophoto...')
    image_url: str



import html
import logging
import urllib.parse as urlparse
from typing import Any, Iterator, List, Optional

import grequests
import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from .. import syndetics
from ..errors import UnstableAPIError
from .query import QueryBuilder
from .types import Book, BookDescription

logger = logging.getLogger('bibliophile')


class _BookPatchedResponse(requests.Response):
    """A response from the BiblioCommons catalog API, with queried book attached.

    This allows us to add metadata to a raw response object, so we know
    exactly which book was the origin of the request.
    """

    _book: Book


class BiblioParser:
    """ Use undocumented BiblioCommons APIs to extract book information. """

    def __init__(
        self,
        biblio_subdomain: str,
        branch: Optional[str] = None,
        isolanguage: Optional[str] = None,
    ):
        self.branch = branch
        self.root = f'https://{biblio_subdomain}.bibliocommons.com/'
        self.isolanguage = isolanguage

    @staticmethod
    def extract_item_id(rss_link: str) -> int:
        """Extract a numeric ID from a link to a book summary.

        'seattle.bibliocommons.com/item/show/2837203030_moby_dick' -> 2837203030

        The RSS feed for a given search links to a page with information about
        that book. The URL is a slug containing an item ID. We need that ID to
        form other requests.
        """
        path = urlparse.urlsplit(rss_link).path
        if not path.startswith('/item/show/'):
            raise UnstableAPIError("RSS link format changed!")

        slug = path.split('/')[-1]  # '2837203030_moby_dick'
        item_id = slug.split('_')[0]  # '2837203030'
        if not item_id.isdigit():
            raise UnstableAPIError("slug format changed!")
        return int(item_id)

    def _async_record(self, book: Book) -> grequests.AsyncRequest:
        """Return an asynchronous request for info about the given book.

        The response content will be the "full record," containing information
        about the given title.
        """
        assert book.full_record_link, "Record URL required to look up item!"
        item_id = self.extract_item_id(book.full_record_link)
        url = urlparse.urljoin(self.root, f'item/full_record/{item_id}')

        # See: https://requests.readthedocs.io/en/latest/user/advanced/#event-hooks
        def attach_book(
            response: _BookPatchedResponse,
            **_kwargs: Any,  # pylint:disable=unused-argument
        ) -> _BookPatchedResponse:
            """ Store the book metadata on the response object. """
            response._book = book  # pylint:disable=protected-access
            return response

        return grequests.get(url, hooks={'response': attach_book})

    def async_book_lookup(self, query: str) -> grequests.AsyncRequest:
        """Formulate & return a request that executes the query.

        The object can be asynchronously queried in a bunch with grequests.
        """
        # BiblioCommons only opens up their API to library employees
        # As of 2011, they said it would publicly-available 'soon'... =(
        rss_search = urlparse.urljoin(self.root, 'search/rss')
        params = {'custom_query': query}

        # grequests doesn't store the (full) url on AsyncRequest objects
        full_url = f"{rss_search}?{urlparse.urlencode(params)}"
        logger.debug("Searching books via RSS: '%s'", full_url)

        return grequests.get(rss_search, params=params)

    @staticmethod
    def book_from_rss_item(rss_item: Tag) -> Book:
        """ Parse out book metadata from XML in an RSS <item>. """
        # The 'description' element contains escaped HTML with <b> labels
        desc_soup = BeautifulSoup(rss_item.description.text, 'html.parser')

        # Try to get call number directly (some libraries don't have it here)
        call_num = desc_soup.find('b', text='Call #:')
        call_number = call_num.next_sibling.strip() if call_num else None

        author_label = desc_soup.find('b', text='Author:')
        author = author_label.find_next('a').text if author_label else None

        desc_label = desc_soup.find('b', text='Description:')
        description = desc_label.find_next('p').text if desc_label else ''

        # Get high-quality cover art from the thumbnail that's given for RSS
        thumbnail = desc_soup.find('div', class_="jacketCoverDiv")
        cover_image: Optional[str] = None
        if thumbnail:
            medium_gif = thumbnail.find('img').attrs['src']
            cover_image = syndetics.higher_quality_cover(image_url=medium_gif)

        return Book(
            full_record_link=rss_item.link.text,  # Can be None!
            title=html.unescape(rss_item.title.text),
            author=author,
            call_number=call_number,
            description=description,
            cover_image=cover_image,
        )

    def matching_books(self, query_response: requests.Response) -> Iterator[Book]:
        """ Yield descriptors of all books matched by the query. """
        soup = BeautifulSoup(query_response.content, 'xml')

        for tag in soup.find('channel').findAll('item'):
            yield self.book_from_rss_item(tag)

    @staticmethod
    def get_call_number(full_record_response: requests.Response) -> str:
        """ Extract a book's call number from its catalog query response. """
        # Yes, that's a JSON endpont that returns HTML. ಠ_ಠ
        soup = BeautifulSoup(full_record_response.json()['html'], 'html.parser')
        call_num = soup.find(testid="callnum_branch").find('span', class_='value')
        text: str = call_num.text
        return text

    def catalog_results(self, books: List[BookDescription]) -> Iterator[Book]:
        """ Yield all books found in the catalog, in no particular order. """

        def grouper(
            # (This will work with any type - could use TypeVar if needed generically)
            input_list: List[BookDescription],
            chunk_size: int,
        ) -> Iterator[List[BookDescription]]:
            for i in range(0, len(input_list), chunk_size):
                yield input_list[i : i + chunk_size]

        # Undocumented, but the API appears to only support lookup of 10 books
        # (We also have a ~900 character search max to contend with)
        query_strings = (
            QueryBuilder.bibliocommons_query(isbn_chunk, self.branch, self.isolanguage)
            for isbn_chunk in grouper(books, 10)
        )
        lookup_requests = [self.async_book_lookup(q) for q in query_strings]
        for response in grequests.imap(lookup_requests):
            for book in self.matching_books(response):
                yield book

    def all_matching_books(self, books: List[BookDescription]) -> Iterator[Book]:
        """ Yield all matching books for the supplied books & branch. """
        search = "Searching library catalog for books"
        if self.branch:
            search += f" at {self.branch}"
        logger.info(search)

        full_record_requests = []

        # First, yield all books with full metadata from the RSS channel
        for book in self.catalog_results(books):
            if book.call_number:
                yield book
            elif not book.full_record_link:
                logger.warning("No link given for %s, can't get call #", book.title)
            else:  # Some metadata found, but we need to more for the call #
                logger.debug("No call # found for %s, fetching record.", book.title)
                full_record_requests.append(self._async_record(book))

        # Then yield books that need additional lookups to fetch call numbers
        for response in grequests.imap(full_record_requests):
            book = response._book  # pylint: disable=protected-access
            yield book._replace(call_number=self.get_call_number(response))



import typing
from typing import Optional, Type, TypeVar

from ..goodreads.types import Book as GoodreadsBook

# Create a generic variable that can be 'Parent', or any subclass.
T = TypeVar('T', bound='BookDescription')


class BookDescription(typing.NamedTuple):
    """This structure contains just the data needed to query the catalog.

    It can be built from a Goodreads shelf result, since that result
    will contain some Goodreads-specific info.
    """

    # Plain ISBN, not ISBN13
    # In rare cases, it's possible for an ISBN to be missing!
    isbn: Optional[str]

    title: str
    author: str

    # NOTE: In Python 3.8 we can use `postponed evaluations of annotations`
    # to just have this method return BookQuery
    @classmethod
    def from_goodreads_book(cls: Type[T], book: GoodreadsBook) -> T:
        return cls(
            isbn=book.isbn,
            title=book.title,
            author=book.author,
        )


class Book(typing.NamedTuple):
    title: str
    author: str
    description: str
    call_number: str
    cover_image: Optional[str]
    full_record_link: Optional[str]



import unittest
import uuid
from unittest import mock

import grequests
from bs4 import BeautifulSoup

from bibliophile.bibliocommons.parse import BiblioParser
from bibliophile.bibliocommons.types import Book as BiblioCommonsBook
from bibliophile.errors import UnstableAPIError


class ItemExtractionTest(unittest.TestCase):
    def test_extract_item_id(self):
        self.assertEqual(
            BiblioParser.extract_item_id(
                'https://seattle.bibliocommons.com/item/show/2837203030_moby_dick',
            ),
            2837203030,
        )

    def test_unstable_link_format(self):
        with self.assertRaises(UnstableAPIError):
            BiblioParser.extract_item_id(
                'https://seattle.bibliocommons.com/item/2837203030_moby_dick',
            )

    def test_unstable_slug_format(self):
        with self.assertRaises(UnstableAPIError):
            BiblioParser.extract_item_id(
                f'https://seattle.bibliocommons.com/item/show/{uuid.uuid4()}'
            )

        with self.assertRaises(UnstableAPIError):
            BiblioParser.extract_item_id(
                'https://seattle.bibliocommons.com/item/show/123/not_expected'
            )


class BookFromRSSTest(unittest.TestCase):
    def test_book_from_rss_item(self):
        rss_item_xml = '''
            <item>
            <title>The Future of the Internet and How to Stop It</title>
            <link>https://sfpl.bibliocommons.com/item/show/2156713093_the_future_of_the_internet_and_how_to_stop_it</link>
            <description>&lt;b&gt;Title:&lt;/b&gt; The Future of the Internet and How to Stop It&lt;br/&gt;
            &lt;b&gt;Author:&lt;/b&gt; &lt;a href="https://sfpl.bibliocommons.com/search?q=%22Zittrain%2C+Jonathan%22&amp;amp;search_category=author&amp;amp;t=author" target="_parent" testid="author_search"&gt;Zittrain, Jonathan&lt;/a&gt;&lt;br/&gt;
            &lt;b&gt;Imprint:&lt;/b&gt; New Haven [Conn.] : &lt;span class="publisher"&gt;Yale University Press&lt;/span&gt;&lt;br/&gt;
            &lt;b&gt;Call #:&lt;/b&gt; 004.678 Z695f&lt;br/&gt;
            &lt;b&gt;Format:&lt;/b&gt; Book&lt;br/&gt;
            &lt;b&gt;Series:&lt;/b&gt; &lt;br/&gt;
            &lt;b&gt;Description:&lt;/b&gt; &lt;p&gt;&lt;/p&gt;&lt;br/&gt;
                            &lt;div class='jacketCoverDiv'&gt;&lt;img alt="The Future of the Internet and How to Stop It" class="jacketCover medium" data-js="" fallback="//cor-cdn-static.bibliocommons.com/assets/default_covers/icon-book-93409e4decdf10c55296c91a97ac2653.png" id="2156713093" src="https://secure.syndetics.com/index.aspx?isbn=9780300124873/SC.GIF&amp;amp;client=sfpl&amp;amp;type=xw12&amp;amp;oclc=" title="The Future of the Internet and How to Stop It" /&gt;&lt;/div&gt;
            &lt;br/&gt;
            &lt;a href="https://sfpl.bibliocommons.com/item/show/2156713093_the_future_of_the_internet_and_how_to_stop_it"&gt;Read More&lt;/a&gt;
            </description>
            <pubDate>Mon, 30 Jun 2008 14:17:00 -0400</pubDate>
            <dc:date>2008-06-30T14:17:00-04:00</dc:date>
            </item>
        '''
        rss_item = BeautifulSoup(rss_item_xml, 'xml').find('item')
        self.assertEqual(
            BiblioParser.book_from_rss_item(rss_item),
            BiblioCommonsBook(
                full_record_link='https://sfpl.bibliocommons.com/item/show/2156713093_the_future_of_the_internet_and_how_to_stop_it',
                title='The Future of the Internet and How to Stop It',
                author='Zittrain, Jonathan',
                call_number='004.678 Z695f',
                description='',
                cover_image='https://secure.syndetics.com/index.aspx?isbn=9780300124873%2FLC.jpg&client=sfpl&type=xw12',
            ),
        )


# TODO: Lots more coverage needed here.
class BiblioParserTest(unittest.TestCase):
    def test_async_book_lookup(self):
        parser = BiblioParser(biblio_subdomain='sfpl')
        fake_request = mock.Mock(spec=grequests.AsyncRequest)
        with mock.patch.object(grequests, 'get') as grequests_get:
            grequests_get.return_value = fake_request
            ret = parser.async_book_lookup(query='(identifier:(0142437239)')
        self.assertEqual(ret, fake_request)
        grequests_get.assert_called_once_with(
            'https://sfpl.bibliocommons.com/search/rss',
            params={'custom_query': '(identifier:(0142437239)'},
        )

    def test_no_books(self):
        parser = BiblioParser(
            biblio_subdomain='sfpl',
            branch='MAIN',
            isolanguage='eng',
        )
        self.assertEqual(list(parser.all_matching_books([])), [])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bibliophile',
 'bibliophile.bibliocommons',
 'bibliophile.goodreads',
 'bibliophile.tests',
 'bibliophile.tests.bibliocommons',
 'bibliophile.tests.goodreads']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4', 'grequests', 'lxml', 'requests']

setup_kwargs = {
    'name': 'bibliophile',
    'version': '1.0.0',
    'description': 'Find books at your local library',
    'long_description': '[![Build Status](https://travis-ci.com/DavidCain/bibliophile-backend.svg?branch=master)](https://travis-ci.com/DavidCain/bibliophile-backend)\n[![Code Coverage](https://codecov.io/gh/DavidCain/bibliophile-backend/branch/master/graph/badge.svg)](https://codecov.io/gh/DavidCain/bibliophile-backend)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n# Bibliophile backend\nThis is a Python-based tool for finding books at the local library.\n\nIt parses your "to read" list from Goodreads.com and checks which items\nare available at a library of your choosing.\n\n## Looking for a GUI?\nThis repository is all you need to use the tool locally via the command line.\nFor a web-based user interface, see [Bibliophile][bibliophile-repo].\n\n# Can I use this?\nIf you live near one of the ~190 public libraries using the BiblioCommons\nsystem, then running this software should work for you. It relies on\nundocumented APIs, so your mileage may vary.\n\n1. Apply for a [Goodreads Developer Key][goodreads-api].\n2. Obtain your Goodreads user id\n3. [Optional] Set both these values in your `.bashrc`\n\n    ```sh\n    export GOODREADS_USER_ID=123456789\n    export GOODREADS_DEV_KEY=whatever-your-actual-key-is\n    ```\n4. [Install Poetry][poetry-install]\n5. Run the script!\n\n    ```sh\n    make  # One-time setup of dependencies\n    ./lookup.py --biblio seattle  # Set to your own city!\n    ```\n\nMake sure you adhere to the terms of [Goodreads\' API][goodreads-api-terms], and\nhave fun.\n\n## Other options\nYou can choose to show only titles available at your local branch, select titles\nfrom another Goodreads shelf, etc. Pass `--help` to see all options:\n\n```\nusage: lookup.py [-h] [--branch BRANCH] [--shelf SHELF] [--biblio BIBLIO]\n                 [--csv CSV]\n                 [user_id] [dev_key]\n\nSee which books you want to read are available at your local library.\n\npositional arguments:\n  user_id          User\'s ID on Goodreads\n  dev_key          Goodreads developer key. See https://www.goodreads.com/api\n\noptional arguments:\n  -h, --help       show this help message and exit\n  --branch BRANCH  Only show titles available at this branch. e.g. \'Fremont\n                   Branch\'\n  --shelf SHELF    Name of the shelf containing desired books\n  --biblio BIBLIO  subdomain of bibliocommons.com (seattle, vpl, etc.)\n  --csv CSV        Output results to a CSV of this name.\n```\n\n## Cloud-based deployment\nThis may also be deployed as Lambda functions in AWS.\nSee the [Bibliophile README][bibliophile-repo] for instructions.\n\n\n[bibliophile-repo]: https://github.com/DavidCain/bibliophile\n[goodreads-api]: https://www.goodreads.com/api\n[goodreads-api-terms]: https://www.goodreads.com/api/terms\n[poetry-install]: https://python-poetry.org/docs/#installation\n',
    'author': 'David Cain',
    'author_email': 'davidjosephcain@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://biblo.dcain.me',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import unittest
from unittest import mock

import requests

from bibliophile.goodreads import images
from bibliophile.goodreads.shelf import ShelfReader
from bibliophile.goodreads.types import Book as GoodreadsBook


class ShelfTests(unittest.TestCase):
    def test_user_id_required(self):
        with self.assertRaises(ValueError):
            ShelfReader(user_id='', dev_key='123')
        with self.assertRaises(ValueError):
            ShelfReader(user_id=None, dev_key='abc')

    def test_key_required(self):
        with self.assertRaises(ValueError):
            ShelfReader(user_id='123456789', dev_key='')
        with self.assertRaises(ValueError):
            ShelfReader(user_id='123456789', dev_key=None)

    def test_wanted_books(self):
        reader = ShelfReader(
            user_id='123456789',
            dev_key='123abc456DEF000foobar9',
        )
        fake_response = mock.Mock()
        fake_response.content = b'''<?xml version="1.0" encoding="UTF-8"?>
            <GoodreadsResponse>
                <Request>
                <authentication>true</authentication>
                    <key><![CDATA[123abc456DEF000foobar9]]></key>
                <method><![CDATA[review_list]]></method>
                </Request>

                  <shelf exclusive=\'true\' id=\'136523200\' name=\'to-read\' sortable=\'false\'></shelf>

                <reviews start="1" end="1" total="1">
                    <review>
                  <id>3253694510</id>
                    <book>
                  <id type="integer">135479</id>
                  <isbn>0140285601</isbn>
                  <isbn13>9780140285604</isbn13>
                  <text_reviews_count type="integer">123450000</text_reviews_count>
                  <uri>omitted-for-test</uri>
                  <title>Cat\'s Cradle</title>
                  <title_without_series>Cat\'s Cradle</title_without_series>
                  <image_url>https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1327867150l/135479._SX98_.jpg</image_url>
                  <small_image_url>https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1327867150l/135479._SY75_.jpg</small_image_url>
                  <large_image_url/>
                  <link>https://www.goodreads.com/book/show/135479.Cat_s_Cradle</link>
                  <num_pages>0</num_pages>
                  <format>Paperback</format>
                  <edition_information/>
                  <publisher>Fake</publisher>
                  <publication_day/>
                  <publication_year>1782</publication_year>
                  <publication_month/>
                  <average_rating>3.14159</average_rating>
                  <ratings_count>123450000</ratings_count>
                  <description>Totally made up description</description>
                <authors>
                <author>
                <id>99999999</id>
                <name>Kurt Vonnegut Jr.</name>
                <role></role>
                <image_url nophoto=\'false\'>
                <![CDATA[https://images.gr-assets.com/authors/1433582280p5/2778055.jpg]]>
                </image_url>
                <small_image_url nophoto=\'false\'>
                <![CDATA[https://images.gr-assets.com/authors/1433582280p2/2778055.jpg]]>
                </small_image_url>
                <link><![CDATA[https://www.goodreads.com/author/show/2778055.Kurt_Vonnegut_Jr_]]></link>
                <average_rating>3.14159</average_rating>
                <ratings_count>123450000</ratings_count>
                <text_reviews_count>123450000</text_reviews_count>
                </author>
                </authors>
                  <published>1789</published>
                <work>  <id>1621115</id>
                  <uri>omitted-for-test</uri>
                </work></book>

                  <rating>0</rating>
                  <votes>0</votes>
                  <spoiler_flag>false</spoiler_flag>
                  <spoilers_state>none</spoilers_state>

                <shelves>

                    <shelf exclusive=\'true\' id=\'136523200\' name=\'to-read\' review_shelf_id=\'2896197272\' sortable=\'false\'></shelf>

                </shelves>

                  <recommended_for></recommended_for>
                  <recommended_by></recommended_by>
                  <started_at></started_at>
                  <read_at></read_at>
                  <date_added>Sun Mar 29 22:51:53 -0700 2020</date_added>
                  <date_updated>Sun Mar 29 22:51:53 -0700 2020</date_updated>
                  <read_count>0</read_count>
                  <body>
                  </body>

                  <comments_count>0</comments_count>
                    <url><![CDATA[https://www.goodreads.com/review/show/3253694510]]></url>
                  <link><![CDATA[https://www.goodreads.com/review/show/3253694510]]></link>
                    <owned>0</owned>
                </review>
                </reviews>

            </GoodreadsResponse>
            '''

        with mock.patch.object(requests, 'get') as requests_get:
            requests_get.return_value = fake_response
            with mock.patch.object(images, 'higher_quality_cover') as higher_quality:
                higher_quality.return_value = 'https://i.gr-assets.com/foo.jpg'
                all_books = list(reader.wanted_books(shelf='to-read'))

        higher_quality.assert_called_once_with(
            'https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1327867150l/135479._SX98_.jpg'
        )

        self.assertEqual(
            all_books,
            [
                GoodreadsBook(
                    goodreads_id="135479",
                    isbn='0140285601',  # ISBN13: 9780140285604
                    title="Cat's Cradle",
                    author='Kurt Vonnegut Jr.',
                    description='Totally made up description',
                    image_url='https://i.gr-assets.com/foo.jpg',
                )
            ],
        )



import unittest
from unittest import mock

from bibliophile.goodreads.images import higher_quality_cover, logger


class CoverImageTests(unittest.TestCase):
    def test_nophoto_url_unchanged(self):
        """ We handle Goodreads' "nophoto" URL. """
        nophoto = (
            "https://www.goodreads.com/"
            "assets/nophoto/book/111x148-bcc042a9c91a29c1d680899eff700a03.png"
        )
        image_url = higher_quality_cover(nophoto)
        self.assertEqual(image_url, nophoto)

    def test_unrecognized_image_format(self):
        """ If the photo is not recognized, we just return it as-is. """
        unexpected_format_url = 'https://images.gr-assets.com/12345789.jpg'
        with mock.patch.object(logger, 'warning') as log_warning:
            image_url = higher_quality_cover(unexpected_format_url)
        log_warning.assert_called_once()
        self.assertEqual(image_url, unexpected_format_url)

    def test_old_photo_(self):
        # Still works, but not expected in the API
        old_format_url = 'https://images.gr-assets.com/books/1550917827s/1202.jpg'
        self.assertEqual(
            higher_quality_cover(old_format_url),
            'https://images.gr-assets.com/books/1550917827l/1202.jpg',
        )

    # This is what the API currently responds with!
    # - images hosted on CDN
    # - The full-size image is used, but with a resize applied
    def test_small_resized_cdn_photo_enlarged(self):
        self.assertEqual(
            higher_quality_cover(
                'https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1367778804l/17081913._SX98_.jpg'
            ),
            'https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1367778804l/17081913.jpg',
        )

    def test_small_raw_cdn_photo_enlarged(self):
        """ We can produce higher-quality images for a small-sized cover served from the CDN. """
        self.assertEqual(
            higher_quality_cover(
                'https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1367778804l/17081913.jpg'
            ),
            'https://i.gr-assets.com/images/S/compressed.photo.goodreads.com/books/1367778804l/17081913.jpg',
        )

    def test_small_base_photo_enlarged(self):
        """ We can produce higher-quality images for a small-sized cover served from the Goodreads site. """
        self.assertEqual(
            # Note that HTTP (not HTTPS) is used!
            higher_quality_cover(
                'http://compressed.photo.goodreads.com/books/1367778804s/17081913.jpg'
            ),
            'http://compressed.photo.goodreads.com/books/1367778804l/17081913.jpg',
        )



"""
Retrieve books on a Goodreads user's "shelf."
"""

import logging
import urllib.parse as urlparse
from typing import Dict, Iterator, Union

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from . import images
from .types import Book

logger = logging.getLogger('bibliophile')


# Max number of pages which can be read at a time
# See: https://www.goodreads.com/api/index#reviews.list
PER_PAGE: int = 200


class ShelfReader:
    """ Read books from a given user's Goodreads shelves. """

    def __init__(self, user_id: str, dev_key: str):
        if not user_id:
            raise ValueError("User ID required!")
        self.user_id = user_id

        if not dev_key:
            raise ValueError("Dev key required!")
        self.dev_key = dev_key

    @staticmethod
    def _get(path: str, params: Dict[str, Union[str, int]]) -> Tag:
        """ Return BS tag for the response to a given Goodreads API route. """
        endpoint = urlparse.urljoin('https://www.goodreads.com/', path)
        resp = requests.get(endpoint, params=params)
        return BeautifulSoup(resp.content, 'xml').find('GoodreadsResponse')

    @staticmethod
    def book_from_review(review: Tag) -> Book:
        """ Parse a <review> XML tag into a Book. """
        return Book(
            goodreads_id=review.find('id', {'type': 'integer'}).text,
            isbn=review.isbn.text or None,  # Can be blank! e.g. in e-Books
            title=review.title.text,
            author=review.author.find('name').text,
            description=review.description.text,
            image_url=images.higher_quality_cover(review.image_url.text),
        )

    def wanted_books(self, shelf: str) -> Iterator[Book]:
        """ All books that the user wants to read. """
        # See: https://www.goodreads.com/api/index#reviews.list
        logger.info("Fetch books on %s for user %s", shelf, self.user_id)
        body = self._get(
            'review/list',
            {
                'v': 2,
                'id': self.user_id,
                'shelf': shelf,
                'key': self.dev_key,
                # TODO: Paginate for shelves over 200 books
                # For now, we only take the top 200...
                'per_page': PER_PAGE,
            },
        )

        for review_tag in body.find('reviews').findAll('review'):
            yield self.book_from_review(review_tag)
